import os
import lxml.etree as ET
import tqdm
import glob
import re

"""
The functions below are helper functions.
"""

def create_if_absent(directory):
    """
    Function creates a directory if it doesn't exist already.
    """
    if not os.path.exists(directory):
        os.makedirs(directory)
        print("Directory", directory, "created")
    else:
        print("Directory", directory, "already exists")
    return

def get_files_with_extension(directory, extension):
    """
    Gets all files with a specific extension (like .xml) from a specific directory.
    """
    try:
        # Using glob for pattern matching
        return [
            os.path.basename(file) for file in glob.glob(f"{directory}/*.{extension}")
        ]
    except OSError as e:
        print(f"Error: {e}")
        return []

def clean_whitespace(text):
    """Clean unnecessary whitespaces from the text."""
    #Collapse multiple spaces and tabs
    text = re.sub(r"[ \t]+", " ", text)
    #Convert multiple newlines to a double newline for paragraph separation
    text = re.sub(r"(\s*\n\s*)+", "\n\n", text)
    return text.strip()

def extract_dbnl_id(f_name):
    """
    Etract dbnl-id from a filename.
    """
    pattern = r"[a-zA-Z_]{4}\d{3}[a-zA-Z_\d]{4}\d{2}"
    regex = re.compile(pattern)
    match = regex.search(f_name)
    return match[0] if match else None

"""
The functions do the core activity of the script:
they take a dir with dbnl-xml files, extract the actual literary texts, and put them in a new dir.
"""

def find_text_element(root):
    """
    Find and return the <text> element from the XML tree. If the number of text elements is not equal to 1, something fishy is going on and
    we get an error. In this case you should get to the BOTTOM of it since it is VERY unexpected.
    """
    text_elements = root.findall(".//text")
    if len(text_elements) != 1:
        raise ValueError(f"Expected exactly 1 <text> element, but found {len(text_elements)}.")
    return text_elements[0]

def get_text_dbnl(xml_file, id, output_dir):
    """
    Extracts text from a dbnl xml-file and saves it as a .txt file.
    In this setup, notes in the texts (e.g. explanations, references) are not removed!
    """
    # Parse the XML file
    tree = ET.parse(xml_file)
    root = tree.getroot()

    # Find the <text> element and extract main text
    text_element = find_text_element(root)
    main_text = clean_whitespace("".join(text_element.itertext()))

    # Save the combined text to a file
    with open(f"{output_dir}/{id}.txt", "w", encoding="utf-8") as f:
        f.write(main_text)


def dbnl_to_txt(input_dir, output_dir="dbnl_txt_files"):
    """
    Puts the pieces of the pipeline together.
    """
    xml_files = get_files_with_extension(input_dir, "xml")
    create_if_absent(output_dir)
    for f in tqdm.tqdm(xml_files):
        id = extract_dbnl_id(f)
        p = os.path.join(input_dir, f)
        try:
            get_text_dbnl(p, id, output_dir)
        except Exception as e:
            error_message = f"file {id} could not be parsed because of: {e}"
            print(error_message)
    return



import os
import asyncio
from langchain_openai import ChatOpenAI
from tqdm.asyncio import tqdm
from typing import Optional
from pydantic import BaseModel, Field, create_model
from langchain_core.prompts import ChatPromptTemplate
from langchain_text_splitters import RecursiveCharacterTextSplitter


text_splitter = RecursiveCharacterTextSplitter(
    chunk_size=400,
    chunk_overlap=20,
    length_function=len,
    is_separator_regex=False,
)

def create_model_name(phenomenon_of_interest: str) -> str:
    return ''.join(word.capitalize() for word in phenomenon_of_interest.split())

def create_llm_analysis_model(phenomenon_of_interest: str) -> BaseModel:
    """
    This model goes to the LLM. The difference here is that the original_sentence field is not present
    as it would be a waste of resources to let the LLM do that.
    """
    model_name = create_model_name(phenomenon_of_interest)
    return create_model(
        f"{model_name}InText",
        explanation=(str, Field(description=f"Explain whether the sentence contains information about {phenomenon_of_interest}")),
        judgement=(bool, Field(description=f"Whether the sentence contains information about {phenomenon_of_interest}"))
    )


def create_full_analysis_model(phenomenon_of_interest: str) -> BaseModel:
    """
    This model is the same as the llm_analysis_model but with the original sentence field added.
    """
    model_name = create_model_name(phenomenon_of_interest)
    return create_model(
        f"{model_name}InText",
        explanation=(str, Field(description=f"Explain whether the sentence contains information about {phenomenon_of_interest}")),
        judgement=(bool, Field(description=f"Whether the sentence contains information about {phenomenon_of_interest}")),
        original_sentence=(str, Field(description="The original sentence from the document"))
    )

def get_system_prompt(phenomenon_of_interest: str) -> str:
    return f"""I am a Cultural Historian and Literary Scholar interested in  {phenomenon_of_interest}. Your task is to read sentences in Early Modern Dutch and indicate whether, 
    given my research interest, the sentence is relevant to my research. You should provide a clear explanation, a boolean judgement, and details about
    {phenomenon_of_interest} if present."""

async def analyze_sentence(sentence: str, structured_llm, FullAnalysisModel):
    try:
        llm_result = await structured_llm.ainvoke(sentence)
        
        full_result = FullAnalysisModel(
            explanation = llm_result.explanation,
            judgement = llm_result.judgement,
            original_sentence = sentence
        )
        return full_result
    except Exception as e:
        print(f"Error analyzing sentence: {e}")
        print(f"Problematic sentence: {sentence}")
        return None

async def analyze_document(input_file: str, phenomenon_of_interest: str, text_splitter = text_splitter,
                           model = "gpt-4o-mini-2024-07-18"):

    with open(input_file, 'r', encoding='utf-8') as f:
        original_text = f.read()

    sentences = text_splitter.split_text(original_text)

    api_key = os.getenv("OPENAI_API_KEY")
    if not api_key:
        raise ValueError("API key not found. Install python-dotenv, make .env style and save the API key there as: OPENAI_API_KEY=your-api-key-here")                           
    llm = ChatOpenAI(model=model)
    LLMAnalysisModel = create_llm_analysis_model(phenomenon_of_interest)
    FullAnalysisModel = create_full_analysis_model(phenomenon_of_interest)

    llm_structured_output = llm.with_structured_output(LLMAnalysisModel)

    system_prompt = get_system_prompt(phenomenon_of_interest)
    prompt = ChatPromptTemplate.from_messages([("system", system_prompt), ("human", "{input}")])
    structured_llm = prompt | llm_structured_output
    
    tasks = [analyze_sentence(sentence, structured_llm, FullAnalysisModel) for sentence in sentences]
    
    return await tqdm.gather(*tasks)


import docx
from docx.shared import RGBColor
from fuzzywuzzy import fuzz
import matplotlib.pyplot as plt
import tqdm 

""""
The highlight relevant passages function still has flaws. 
I have not been able to come up with a system that has 100% recall and 100% precision,
but theoretically that should be possible.

The barcode plot has the same issue (with a different solution though) but here the problem
is less urgent because the barcode plot is just used to give an indication so there's more
room for error.
"""
import matplotlib.pyplot as plt
import docx
from difflib import SequenceMatcher
import re
import tqdm
from fuzzywuzzy import fuzz
import numpy as np

def highlight_relevant_passages(input_file, output_file, relevant_passages):
    # Read the original text
    with open(input_file, 'r', encoding='utf-8') as file:
        original_text = file.read()

    # Create a new Word document
    doc = docx.Document()

    # Preprocess the text to create an index
    text_index = {}
    for i in range(len(original_text) - 5):
        five_gram = original_text[i:i+5]
        if five_gram not in text_index:
            text_index[five_gram] = []
        text_index[five_gram].append(i)

    # Process the text and add it to the document with highlighting
    current_pos = 0
    for passage in tqdm.tqdm(relevant_passages, desc="Processing passages"):
        # Find the best match for the passage using the index
        best_match = None
        best_ratio = 0
        potential_starts = text_index.get(passage[:5], [])
        
        for start in potential_starts:
            end = start + len(passage)
            substring = original_text[start:end]
            ratio = fuzz.ratio(passage, substring)
            if ratio > best_ratio:
                best_ratio = ratio
                best_match = (start, end)

        if best_match:
            doc.add_paragraph(original_text[current_pos:best_match[0]])

            # Add the matched text with yellow highlighting
            p = doc.add_paragraph()
            run = p.add_run(original_text[best_match[0]:best_match[1]])
            run.font.highlight_color = docx.enum.text.WD_COLOR_INDEX.YELLOW

            current_pos = best_match[1]

    # Add any remaining text
    if current_pos < len(original_text):
        doc.add_paragraph(original_text[current_pos:])

    # Save the document
    doc.save(output_file)

def map_to_original_index(original_text, alphanumeric_index):
    alphanumeric_count = 0
    for i, char in enumerate(original_text):
        if char.isalnum():
            if alphanumeric_count == alphanumeric_index:
                return i
            alphanumeric_count += 1
    return len(original_text)

def create_plain_text_output(output_file, document_name, relevant_passages):
    with open(output_file, 'w', encoding='utf-8') as file:
        file.write(f"In this text: {document_name} the following has been found (numbered):\n")
        for n, passage in enumerate(relevant_passages):
            file.write(f"No. {n}:    {passage}\n")

def create_barcode_plot(relevant_passages, all_passages):
    barcode = np.array([1 if p in relevant_passages else 0 for p in all_passages])
    pixel_per_bar = 4
    dpi = 100
    fig = plt.figure(figsize=(len(barcode) * pixel_per_bar / dpi, 2), dpi=dpi)
    ax = fig.add_axes([0, 0, 1, 1])  # span the whole figure
    ax.set_axis_off()
    ax.imshow(barcode.reshape(1, -1), cmap='binary', aspect='auto',
          interpolation='nearest')
    plt.show()



from setuptools import setup, find_packages

setup(
    name='dbnl_bear',
    version='0.3',
    packages=find_packages(),
    install_requires=["lxml", "tqdm", "python-dotenv", "langchain_core", "langchain_openai", "tqdm",
                      "typing", "langchain_text_splitters", "docx", "fuzzywuzzy", "pydantic",
                     "python-docx", "python-Levenshtein"],
    description='A Python package to use an AI assistant for historical analysis',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    author='Arjan van Dalfsen',
    author_email='j.a.vandalfsen@uu.nl',
    url='https://github.com/trister95/dbnl_bear',
    classifiers=[
        'Programming Language :: Python :: 3',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.10',
)
